package com.example.dollarstroke;

import java.io.IOException;
import java.io.InputStream;
import java.io.File;
import java.io.FileOutputStream;

import org.xmlpull.v1.XmlPullParserException;

import android.widget.ImageView;
import android.util.Log;
import android.content.Context;
import android.content.res.TypedArray;
import android.graphics.Canvas;
import android.graphics.Color;
import android.view.MotionEvent;
import android.graphics.Paint;
import android.graphics.Bitmap;
import android.graphics.drawable.Drawable;
import android.text.TextPaint;
import android.util.AttributeSet;
import android.graphics.Path;
import android.view.inputmethod.InputConnection;
import android.inputmethodservice.InputMethodService;
import android.view.KeyEvent;
import android.graphics.DashPathEffect;
import android.widget.TextView;

import java.util.ArrayList;

import android.os.Environment;

/**
 * TODO: document your custom view class.
 */
public class DollarStrokeView extends ImageView {
	
	private String mExampleString = null;  
	private int mExampleColor = Color.RED; 
	private float mExampleDimension = 0;   
	private Drawable mExampleDrawable;
	private DashPathEffect mDashedEffect = new DashPathEffect(new float[] {10,20}, 0);

	private TextPaint mTextPaint;
	private float mTextWidth;
	private float mTextHeight;
	private Paint mPaint;
	private Bitmap mBitmap;
	private Canvas mCanvas;  
	private float mCurX;
	private float mCurY;
	private int mNumValX;

	private int mUseDirection = DollarRecognizer.RECON_FLAG_USE_ENDPOINTDIST | DollarRecognizer.RECON_FLAG_USE_ANGLEDIR ;
	private boolean mIsSplitPane = false;
	private boolean mIsNumLock = false;
	private boolean mIsCapLock = false;

	private Path mPath = null;

	private int mStrokeCount = 0;
	private DollarRecognizer.Recognizer pAlphaDollarRecon = null;
	private DollarRecognizer.Recognizer pNumericDollarRecon = null;
	private ArrayList<DollarRecognizer.Point> mPoints = null;

	InputMethodService mInputMethodService = null;

	public DollarStrokeView(Context context) {
		super(context);
		init(null, 0);
	}

	public DollarStrokeView(Context context, AttributeSet attrs) {
		super(context, attrs);
		init(attrs, 0);
	}

	public DollarStrokeView(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs, defStyle);
		init(attrs, defStyle);
	}

	private void init(AttributeSet attrs, int defStyle) {

		// Load attributes
		final TypedArray a = getContext().obtainStyledAttributes(attrs,
				R.styleable.DollarStrokeView, defStyle, 0);

		mExampleString = a
				.getString(R.styleable.DollarStrokeView_exampleString);
		mExampleColor = a.getColor(R.styleable.DollarStrokeView_exampleColor,
				mExampleColor);
		// Use getDimensionPixelSize or getDimensionPixelOffset when dealing
		// with
		// values that should fall on pixel boundaries.
		mExampleDimension = a.getDimension(
				R.styleable.DollarStrokeView_exampleDimension,
				mExampleDimension);

		if (a.hasValue(R.styleable.DollarStrokeView_exampleDrawable)) {
			mExampleDrawable = a
					.getDrawable(R.styleable.DollarStrokeView_exampleDrawable);
			mExampleDrawable.setCallback(this);
		}

		a.recycle();

		setFocusable(true);

		// Set up a default TextPaint object
		mTextPaint = new TextPaint();
		mTextPaint.setFlags(Paint.ANTI_ALIAS_FLAG);
		mTextPaint.setTextAlign(Paint.Align.LEFT);
		
		mPaint = new Paint();
		mPaint.setAntiAlias(true);
		mPaint.setARGB(255, 255, 255, 255);
        
		mPath   = new Path();
		mPoints = new ArrayList<DollarRecognizer.Point>();

		InputStream inputStream;
		try {
			if(!isInEditMode()){
				inputStream = getContext().getAssets().open("alphaPointGestureDb.xml");
				pAlphaDollarRecon = new DollarRecognizer.PRecognizer(inputStream);
			}

		} catch( XmlPullParserException e ) {
			Log.d("DollarStrokeView", e.getMessage()) ;
			e.printStackTrace();
		} catch( IOException e ) {
			Log.d("DollarStrokeView", e.getMessage());
			e.printStackTrace();
		}

		try {
			if(!isInEditMode()){
				inputStream = getContext().getAssets().open("numericPointGestureDb.xml");
				pNumericDollarRecon = new DollarRecognizer.PRecognizer(inputStream);
			}

		} catch( XmlPullParserException e ) {
			Log.d("DollarStrokeView", e.getMessage());
			e.printStackTrace();
		} catch( IOException e ) {
			Log.d("DollarStrokeView", e.getMessage());
			e.printStackTrace();
		}

		// Update TextPaint and text measurements from attributes
		invalidateTextPaintAndMeasurements();
	}

	private void invalidateTextPaintAndMeasurements() {
		mTextPaint.setTextSize(mExampleDimension);
		mTextPaint.setColor(mExampleColor);
		if( mExampleString != null )
			mTextWidth = mTextPaint.measureText(mExampleString);

		Paint.FontMetrics fontMetrics = mTextPaint.getFontMetrics();
		mTextHeight = fontMetrics.bottom;
	}

	@Override
	protected void onDraw(Canvas canvas)
	{
		super.onDraw(canvas);
/*
		if( mBitmap != null ) 
		{
			canvas.drawBitmap(mBitmap, 0, 0, null);
			canvas.drawPath(mPath, mPaint);
		}
*/
		// TODO: consider storing these as member variables to reduce
		// allocations per draw cycle.
		int paddingLeft   = getPaddingLeft();
		int paddingTop    = getPaddingTop();
		int paddingRight  = getPaddingRight();
		int paddingBottom = getPaddingBottom();

		int contentWidth  = getWidth()  - paddingLeft - paddingRight;
		int contentHeight = getHeight() - paddingTop  - paddingBottom;

		if( !mPath.isEmpty() )
		{
			mPaint.setColor(Color.BLUE);
			mPaint.setStrokeWidth(5);
			mPaint.setStyle(Paint.Style.STROKE);
			canvas.drawPath(mPath, mPaint);
		}

		if( mIsSplitPane )
		{
			mNumValX = contentWidth - (contentWidth / 2);
			mPaint.setColor(mExampleColor);
			mPaint.setStrokeWidth(2);
			mPaint.setStyle(Paint.Style.STROKE);
			mPaint.setPathEffect(mDashedEffect);
			canvas.drawLine(mNumValX, 0, mNumValX, contentHeight, mPaint);
			canvas.drawText("Numeric", mNumValX+20, paddingTop+20, mTextPaint);
		}

		String numString = mIsNumLock ? "Numeric" : mIsCapLock ? "Alphabet - Cap Lock" : "Alphabet";
		canvas.drawText(numString, paddingLeft+20, paddingTop+20, mTextPaint);

		if( mExampleString != null )
		{
			/*
			// Draw the text.
			canvas.drawText(mExampleString, paddingLeft
					+ (contentWidth - mTextWidth) / 2, paddingTop
					+ (contentHeight + mTextHeight) / 2, mTextPaint);

			// Draw the example drawable on top of the text.
			if( mExampleDrawable != null )
			{
				mExampleDrawable.setBounds(paddingLeft, paddingTop, paddingLeft
						+ contentWidth, paddingTop + contentHeight);
				mExampleDrawable.draw(canvas);
			}
			*/
		}
	}
	
	@Override protected void onSizeChanged(int w, int h, int oldw, int oldh) 
	{
		super.onSizeChanged(w,h,oldw,oldh);
/*
		int curW = mBitmap != null ? mBitmap.getWidth() : 0;
		int curH = mBitmap != null ? mBitmap.getHeight() : 0;
		if (curW >= w && curH >= h) {
		    return;
		}

		if (curW < w) curW = w;
		if (curH < h) curH = h;
		
		Bitmap newBitmap = Bitmap.createBitmap(
					w, 
					h,
					Bitmap.Config.RGB_565
			);
		
		Canvas newCanvas = new Canvas();
		newCanvas.setBitmap(newBitmap);
		
		if( mBitmap != null ) {
			newCanvas.drawBitmap(mBitmap, 0, 0, null);
		}

		mBitmap = newBitmap;
		mCanvas = newCanvas;
*/
	}
	
	/**
	 * Gets the example string attribute value.
	 * 
	 * @return The example string attribute value.
	 */
	public String getExampleString() {
		return mExampleString;
	}

	/**
	 * Sets the view's example string attribute value. In the example view, this
	 * string is the text to draw.
	 * 
	 * @param exampleString
	 *            The example string attribute value to use.
	 */
	public void setExampleString(String exampleString) {
		mExampleString = exampleString;
		invalidateTextPaintAndMeasurements();
	}

	/**
	 * Gets the example color attribute value.
	 * 
	 * @return The example color attribute value.
	 */
	public int getExampleColor() {
		return mExampleColor;
	}

	/**
	 * Sets the view's example color attribute value. In the example view, this
	 * color is the font color.
	 * 
	 * @param exampleColor
	 *            The example color attribute value to use.
	 */
	public void setExampleColor(int exampleColor) {
		mExampleColor = exampleColor;
		invalidateTextPaintAndMeasurements();
	}

	/**
	 * Gets the example dimension attribute value.
	 * 
	 * @return The example dimension attribute value.
	 */
	public float getExampleDimension() {
		return mExampleDimension;
	}

	/**
	 * Sets the view's example dimension attribute value. In the example view,
	 * this dimension is the font size.
	 * 
	 * @param exampleDimension
	 *            The example dimension attribute value to use.
	 */
	public void setExampleDimension(float exampleDimension) {
		mExampleDimension = exampleDimension;
		invalidateTextPaintAndMeasurements();
	}

	/**
	 * Gets the example drawable attribute value.
	 * 
	 * @return The example drawable attribute value.
	 */
	public Drawable getExampleDrawable() {
		return mExampleDrawable;
	}

	/**
	 * Sets the view's example drawable attribute value. In the example view,
	 * this drawable is drawn above the text.
	 * 
	 * @param exampleDrawable
	 *            The example drawable attribute value to use.
	 */
	public void setExampleDrawable(Drawable exampleDrawable) {
		mExampleDrawable = exampleDrawable;
	}      

	public DollarRecognizer.PointTemplate writeStrokeSequence(String name)
	{
		DollarRecognizer.PointTemplate pntTmpl = null;

		if( mPoints.size() > 1 && name != null )
		{
			File fileDir = getContext().getExternalFilesDir(null);
			File file    = new File(fileDir,"userGesture.xml");

			try {
				//if file doesn't exists, then create it
				if( !file.exists() )
				{
					file.createNewFile();
				}

				FileOutputStream outputStream = new FileOutputStream(file,true);
				pntTmpl = new DollarRecognizer.PointTemplate(name,mPoints,20);
				DollarRecognizer.writePointsToXml(outputStream, pntTmpl);
				outPutViewInfo(String.format("Wrote Gesture Sequence name=%s\n",name), true);
			} catch (Exception e) {
				Log.d("DollarStrokeView", e.getMessage() + " " + file);
			}
		}
		return pntTmpl;
	}

	public void addStrokeSequence(String name)
	{
		DollarRecognizer.PointTemplate pntTmpl = writeStrokeSequence(name);
		if( pntTmpl != null )
		{
			pAlphaDollarRecon.addPointTemplate(pntTmpl);
		}
	}

	public void setInputMethodService(InputMethodService ims){ mInputMethodService = ims; }

	public void setPaneSplit(boolean flag){ mIsSplitPane = flag; invalidate(); }

	public void setNumLock(boolean flag)
	{
		mIsNumLock = flag;
		invalidate();
	}

	public void setCapLock(boolean flag)
	{
		mIsCapLock = flag;
		invalidate();
	}

	public void recognizeStrokes()
	{
		if( mPoints.size() > 1 )
		{
			DollarRecognizer.Recognizer pDolRecon;
			
			if( mIsNumLock || (mIsSplitPane && mPoints.get(0).X > mNumValX) ) 
			{
				pDolRecon = pNumericDollarRecon;
			} else {
				pDolRecon = pAlphaDollarRecon;
			}

			if( pDolRecon == null )
			{
				Log.d("DollarStrokeView", "Recognizer not created");
				return;
			}

			DollarRecognizer.Results results 
				= pDolRecon.Recognize(mPoints,mUseDirection);
			if( results.mPntTemplate == null )
			{
				Log.d("DollarStrokeView", "No Match Found");
				outPutViewInfo("No Match Found",true);
				return;
			}

			int asciNum = results.mPntTemplate.mNumValue;
			Log.d("DollarStrokeView", "Found Match: " + asciNum + " : " + results.mPntTemplate.mName);
			// handleStrokeInput(asciNum);
			outPutViewInfo( String.format(
				"%s score=%f directionAngle=%f\n",
				results.mPntTemplate.mName,results.mScore, results.angle),true);
		}
	}

	@Override
	public boolean onTouchEvent(MotionEvent event)
	{
		switch( event.getAction() )
		{
			case MotionEvent.ACTION_DOWN:
				mStrokeCount++;
				mPath.reset();
				mPoints.clear();
				collectPathPoints(event);
				drawPoint(mPath);
				break;

			case MotionEvent.ACTION_MOVE:
				collectPathPoints(event);
				drawPoint(mPath);
				break;

			case MotionEvent.ACTION_UP:
				collectPathPoints(event);
				drawPoint(mPath);
				recognizeStrokes();
				break;
		}
		return true;
	}

	private void collectPathPoints(MotionEvent event)
	{
		int N = event.getHistorySize();
		int P = event.getPointerCount();

		for (int i = 0; i < N; i++) {
			for (int j = 0; j < P; j++) {
				mCurX = event.getHistoricalX(j, i);
				mCurY = event.getHistoricalY(j, i);
				outPutViewInfo(String.format("%d) X=%3.0f Y=%3.0f\n", mStrokeCount,mCurX,mCurY), showCoord);
				mPoints.add(new DollarRecognizer.Point(mCurX,mCurY,mStrokeCount));
				if( mPath.isEmpty() ) {
					mPath.moveTo(mCurX,mCurY);
				} else {
					mPath.lineTo(mCurX,mCurY);
				}
			  }
		}

		for (int j = 0; j < P; j++) 
		{
			mCurX = event.getX(j);
			mCurY = event.getY(j);
			outPutViewInfo(String.format("%d) X=%3.0f Y=%3.0f\n", mStrokeCount,mCurX,mCurY), showCoord);
			if( mPath.isEmpty() ) {
				mPath.moveTo(mCurX,mCurY);
			} else {
				mPath.lineTo(mCurX,mCurY);
			}
		}
	}

	private void drawPoint(Path path)
	{
		mPaint.setColor(Color.BLUE);
		mPaint.setStrokeWidth(5);
		mPaint.setStyle(Paint.Style.STROKE);
		mPaint.setPathEffect(null);
//		mCanvas.drawPath(path, mPaint);
		invalidate();
	}

	public void handleStrokeInput(int asciNum)
	{
		InputConnection ic = mInputMethodService.getCurrentInputConnection();
		switch( asciNum )
		{
			case 8 :
				ic.sendKeyEvent(new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_DEL));
				ic.sendKeyEvent(new KeyEvent(KeyEvent.ACTION_UP,   KeyEvent.KEYCODE_DEL));
				return;
			case 13 :
				ic.sendKeyEvent(new KeyEvent(KeyEvent.ACTION_DOWN, KeyEvent.KEYCODE_ENTER));
				ic.sendKeyEvent(new KeyEvent(KeyEvent.ACTION_UP,   KeyEvent.KEYCODE_ENTER));
				return;
			default:
				break;
		}

		if( asciNum > 63 && asciNum < 91 )
		{
			if( mIsCapLock ) {
				ic.commitText(String.valueOf((char)asciNum).toUpperCase(), 1);
			} else {
				ic.commitText(String.valueOf((char)asciNum).toLowerCase(), 1);
			}
			return;
		}

		if( asciNum > 31 && asciNum < 127 )
		{
			ic.commitText(String.valueOf((char)asciNum), 1);
			return;
		}
	}

	private TextView mInfoText = null;
	boolean showCoord = false;

        public void clearPointHistory()
        {
                mStrokeCount = 0;
                mPath.reset();
                mPoints.clear();
        }

        public void clear()
        {
                mStrokeCount = 0;
                mPath.reset();
                mPoints.clear();

                if( mInfoText != null )
                        mInfoText.setText("");
		invalidate();
        }

	public void setOutputTextView(TextView txtV){ mInfoText = txtV; }

	public void turnOffXY(boolean flag){ showCoord = flag; }

	public void useDirection(boolean flag){ mUseDirection = mUseDirection; }

	private void outPutViewInfo(String output, boolean flagOn )
	{
		if( mInfoText != null && flagOn )
		{
			mInfoText.append(output);
		}
	}
}
